package builtins

import (
	"context"
	"errors"
	"strings"
	"ucl.lmika.dev/ucl"
)

func Strs() ucl.Module {
	return ucl.Module{
		Name: "strs",
		Builtins: map[string]ucl.BuiltinHandler{
			"to-upper":   toUpper,
			"to-lower":   toLower,
			"trim":       trim,
			"split":      split,
			"join":       join,
			"has-prefix": hasPrefix,
		},
	}
}

func toUpper(ctx context.Context, args ucl.CallArgs) (any, error) {
	var s string
	if err := args.Bind(&s); err != nil {
		return nil, err
	}

	return strings.ToUpper(s), nil
}

func toLower(ctx context.Context, args ucl.CallArgs) (any, error) {
	var s string
	if err := args.Bind(&s); err != nil {
		return nil, err
	}

	return strings.ToLower(s), nil
}

func trim(ctx context.Context, args ucl.CallArgs) (any, error) {
	var s string
	if err := args.Bind(&s); err != nil {
		return nil, err
	}

	return strings.TrimSpace(s), nil
}

func hasPrefix(ctx context.Context, args ucl.CallArgs) (any, error) {
	var s, prefix string
	if err := args.Bind(&s, &prefix); err != nil {
		return nil, err
	}

	return strings.HasPrefix(s, prefix), nil
}

func split(ctx context.Context, args ucl.CallArgs) (any, error) {
	var s string
	if err := args.Bind(&s); err != nil {
		return nil, err
	}

	sep := ""
	if args.NArgs() > 0 {
		if err := args.Bind(&sep); err != nil {
			return nil, err
		}
	}

	n := -1
	if args.HasSwitch("max") {
		if err := args.BindSwitch("max", &n); err != nil {
			return nil, err
		}
	}

	return ucl.StringListObject(strings.SplitN(s, sep, n)), nil
}

func join(ctx context.Context, args ucl.CallArgs) (any, error) {
	var (
		what ucl.Object
		tok  string
	)

	if err := args.Bind(&what); err != nil {
		return nil, err
	}

	if args.NArgs() > 0 {
		if err := args.Bind(&tok); err != nil {
			return nil, err
		}
	}

	switch t := what.(type) {
	case ucl.Listable:
		var sb strings.Builder
		for i := 0; i < t.Len(); i++ {
			if i > 0 {
				sb.WriteString(tok)
			}
			sb.WriteString(t.Index(i).String())
		}
		return sb.String(), nil
	case ucl.Iterable:
		first := true

		var sb strings.Builder
		for t.HasNext() {
			v, err := t.Next(ctx)
			if err != nil {
				return nil, err
			}
			if !first {
				sb.WriteString(tok)
			} else {
				first = false
			}
			sb.WriteString(v.String())
		}

		return sb.String(), nil
	}

	return nil, errors.New("expected listable or iterable as arg 1")
}
